//
// nono
// Copyright (C) 2024 nono project
// Licensed under nono-license.txt
//

//
// メインバス (virt68k)
//

// IODevice
//   +-- MainbusBaseDevice (InitMainbus() と FC アクセスを持つ)
//   |    +-- MainbusDevice (これがメインバス、システムに1つ)
//   |    |    +-- Mainbus24Device (上位8ビットがテーブルで表せるメインバス)
//   |    |    |    +-- LunaMainbus
//   |    |    |    |    +-- Luna1Mainbus
//   |    |    |    |    +-- Luna88kMainbus
//   |    |    |    +-- NewsMainbus
//   |    |    |    |
//   |    |    |    |  +----------------+
//   |    |    |    +--| Virt68kMainbus |
//   |    |    |       +----------------+
//   |    |    |
//   |    |    +-- X68kMainbus
//   |    +-- X68kIODevice
//   +-- XPbusDevice

#include "mainbus_virt68k.h"
#include "interrupt.h"
#include "mainapp.h"
#include "mainram.h"
#include "v68kio.h"

// コンストラクタ
Virt68kMainbus::Virt68kMainbus()
{
	NEWDV(BusErr, new BusErrDevice());
	NEWDV(Interrupt, new Virt68kInterrupt());
	NEWDV(MainRAM, new MainRAMDevice());
	NEWDV(V68kIO, new V68kIODevice());

	// $ff'000000 以降がデバイス領域。
	// メインメモリは Init() で設定する。
	for (int i = 0; i < devtable.size() - 1; i++) {
		devtable[i] = pBusErr.get();
	}
	devtable[0xff] = pV68kIO.get();

	// 2MB ずつ 32bit 全域。
	acc.resize(2048);
	monitor_accstat.SetSize(80, 2 + 32);
}

// デストラクタ
Virt68kMainbus::~Virt68kMainbus()
{
}

// メインバスの初期化
bool
Virt68kMainbus::InitMainbus()
{
	// RAM は容量が確定した後のここで配置。
	// 16MB 単位になっている。
	auto mainram = dynamic_cast<MainRAMDevice *>(pMainRAM.get());
	int ramsize_MB = mainram->GetSizeMB();
	for (int i = 0; i < ramsize_MB / 16; i++) {
		devtable[i] = mainram;
	}

	// アクセス状況用のマップを作成。
	InitAccStat();

	return true;
}

// ブートページを切り替える。
void
Virt68kMainbus::SwitchBootPage(bool isram)
{
	// まだサポートしていない。
}
