//
// nono
// Copyright (C) 2024 nono project
// Licensed under nono-license.txt
//

//
// VirtIO ネットワーク
//

#include "virtio_net.h"
#include "virtio_def.h"
#include "config.h"
#include "ethernet.h"
#include "macaddr.h"
#include "memorystream.h"
#include "scheduler.h"

class VirtIONetReq : public VirtIOReq
{
 public:
	virtio_net_hdr hdr;
};

// デバイス構成レイアウト
class VirtIONetConfigWriter
{
 public:
	std::array<u8, 6> mac {};
	le16 status {};
	le16 max_virtqueue_pairs {};
	le16 mtu {};

 public:
	void WriteTo(uint8 *dst) const;
};

// コンストラクタ
VirtIONetDevice::VirtIONetDevice(int slot_)
	: inherited(OBJ_VIRTIO_NET, slot_)
{
	// 短縮形
	AddAlias("VNet");
	AddAlias("Ethernet");

	// 割り込み名
	strlcpy(intrname, "VIONet", sizeof(intrname));
	// 完了通知メッセージ
	msgid = MessageID::VIRTIO_NET_DONE;

	device_id = VirtIO::DEVICE_ID_NET;
	vqueues.emplace_back(0, "ReceiveQ1", 16);
	vqueues.emplace_back(1, "TransmitQ1", 8);

	// time は都度設定する
	event.func = ToEventCallback(&VirtIONetDevice::RxEvent);
	event.Regist("VirtIONet RX");

	// モニタの行数。
	int vqlines = 0;
	for (const auto& q : vqueues) {
		vqlines += 7 + q.num_max;
	}

	monitor.func = ToMonitorCallback(&VirtIONetDevice::MonitorUpdate);
	monitor.SetSize(MONITOR_WIDTH, 3 + vqlines);
	monitor.Regist(ID_MONITOR_VIRTIO_NET);
}

// デストラクタ
VirtIONetDevice::~VirtIONetDevice()
{
	// EthernetDevice と同じ…。
	if ((bool)hostnet) {
		hostnet->SetRxCallback(NULL);
	}
}

// 動的なコンストラクション
bool
VirtIONetDevice::Create()
{
	if (inherited::Create() == false) {
		return false;
	}

	// EthernetDevice とほぼ同じ…。
	hostnet.reset(new HostNetDevice(this, 0));

	hostnet->SetRxCallback(ToDeviceCallback(&VirtIONetDevice::HostRxCallback));

	return true;
}

// 初期化
bool
VirtIONetDevice::Init()
{
	if (inherited::Init() == false) {
		return false;
	}

	// MAC アドレスを取得。
	macaddr_t macaddr;
	if (EthernetDevice::GetConfigMacAddr(0, &macaddr, false) == false) {
		// エラーメッセージ表示済み。
		return false;
	}
	// この先設定するタイミングがないのでここで設定する?
	hostnet->SetMyAddr(macaddr);
	// 表示用。
	macaddr_str = macaddr.ToString(':');

	// DEVICE_FEATURES と構成レイアウトを用意。
	VirtIONetConfigWriter cfg;
	SetDeviceFeatures(VIRTIO_NET_F_CSUM);
	SetDeviceFeatures(VIRTIO_NET_F_GUEST_CSUM);
	SetDeviceFeatures(VIRTIO_NET_F_MTU), cfg.mac = macaddr;
	SetDeviceFeatures(VIRTIO_NET_F_MAC), cfg.mtu = 1500;
	SetDeviceFeatures(VIRTIO_NET_F_MRG_RXBUF);
	cfg.WriteTo(&device_config[0]);

	// ホストからの受信コールバックを登録
	scheduler->ConnectMessage(MessageID::HOSTNET_RX(0), this,
		ToMessageCallback(&VirtIONetDevice::RxMessage));

	return true;
}

void
VirtIONetDevice::MonitorUpdate(Monitor *, TextScreen& screen)
{
	int y = 0;

	screen.Clear();

	y = MonitorUpdateDev(screen, y);

	screen.Puts(0, y, "MACAddress:");
	screen.Puts(12, y, macaddr_str.c_str());
	y++;

	y++;
	for (const auto& q : vqueues) {
		y = MonitorUpdateVirtQueue(screen, y, q);
		y++;
		y = MonitorUpdateVirtQDesc(screen, y, q);
		y++;
	}
}

void
VirtIONetDevice::QueueReady()
{
	if (vq->idx == 0) {
		// ここで受信開始だろうか。
		hostnet->EnableRx(true);
	}
}

// ディスクリプタを一つ処理する。
void
VirtIONetDevice::ProcessDesc(VirtQueue *q)
{
	VirtIONetReq req;
	VirtQDesc desc;
	uint32 desc_idx;
	uint32 reqlen = 0;

	if (q->idx == 0) {
		putlog(0, "QUEUE_NOTIFY on rx queue (VQ%u)?", q->idx);
		return;
	}

	// 1チャンク目は先頭 12 バイトが Read-Only なヘッダのはず。
	desc_idx = ReadLE16(q->driver + 4 + (q->last_avail_idx % q->num) * 2);
	reqlen += ReadDesc(q, &desc, desc_idx);
	if (__predict_false(desc.IsRead() == false)) {
		putlog(0, "NetHdr segment must be device-readable");
		return;
	}
	// 仕様上は1バイト単位でチェインされてもいいのだが、面倒なので無視。
	if (desc.len < 12) {
		putlog(0, "NetHdr segment too short (desc.len=$%x)", desc.len);
		return;
	}
	req.idx = desc_idx;
	req.hdr.flags       = ReadU8(desc.addr + 0);
	req.hdr.gso_type    = ReadU8(desc.addr + 1);
	req.hdr.hdr_len     = ReadLE16(desc.addr + 2);
	req.hdr.gso_size    = ReadLE16(desc.addr + 4);
	req.hdr.csum_start  = ReadLE16(desc.addr + 6);
	req.hdr.csum_offset = ReadLE16(desc.addr + 8);
	req.hdr.num_buffers = ReadLE16(desc.addr + 10);
	putlog(2, "req.hdr $%08x: flags=%02x gso=%02x hdr_len=%04x gso_size=%04x "
		"csum_start=%04x offset=%04x num=%04x",
		desc.addr,
		req.hdr.flags,
		req.hdr.gso_type,
		req.hdr.hdr_len,
		req.hdr.gso_size,
		req.hdr.csum_start,
		req.hdr.csum_offset,
		req.hdr.num_buffers);

	// 同一セグメント内にデータ(ReadOnly) が連続していても構わない。
	const auto HDRSZ = sizeof(req.hdr);
	if (desc.len > HDRSZ) {
		AddDataSegment(&req, desc.addr + HDRSZ, desc.len - HDRSZ,
			" (header segment)");
	}

	while (desc.IsNext()) {
		desc_idx = desc.next;
		reqlen += ReadDesc(q, &desc, desc_idx);
		AddDataSegment(&req, desc.addr, desc.len, NULL);
	}
	putlog(3, "reqlen=%08x", reqlen);

	NetPacket tx_packet;
	for (const auto seg : req.buf) {
		uint32 addr = seg.addr;
		for (uint32 end = addr + seg.len; addr < end; ) {
			tx_packet.Append(ReadU8(addr++));
		}
	}
	hostnet->Tx(tx_packet);

	CommitDesc(q, req.idx, reqlen);
}

// これは Host スレッドから呼ばれる
void
VirtIONetDevice::HostRxCallback()
{
	// スレッドを超えるためにメッセージを投げる
	scheduler->SendMessage(MessageID::HOSTNET_RX(0));
}

// パケット受信通知
void
VirtIONetDevice::RxMessage(MessageID msgid_, uint32 arg)
{
	// 受信イベントが止まっていれば動かす。
	// 受信イベントがすでに動いていれば、1パケット受信完了後に
	// ホストキューを確認するので、ここでは何もしなくてよい。
	if (event.IsRunning() == false) {
		event.time = 0;
		scheduler->StartEvent(event);
	}
}

// 受信ループイベント。
void
VirtIONetDevice::RxEvent(Event& ev)
{
	VirtQueue *q = &vqueues[0];

	// 空きディスクリプタがなければ、空くまでポーリングで待ち続ける。
	uint16 avail_idx = ReadLE16(q->driver + 2);
	if (q->last_avail_idx == avail_idx) {
		event.time = 500_usec;
		putlog(2, "%s: wait %u usec", __func__, (uint)(event.time / 1_usec));
		scheduler->RestartEvent(event);
		return;
	}

	// ディスクリプタに空きが出来たので、ホストキューから1パケット取り出す。
	assert(rx_packet.length == 0);
	if (hostnet->Rx(&rx_packet) == false) {
		// 取り出せなくなったら、ここでイベントループを終了。
		rx_packet.Clear();
		putlog(2, "Rx no more rx_packet");
		return;
	}

	putlog(2, "%s: %u bytes from host queue", __func__, rx_packet.length);

	// rx_packet が埋まって、空きディスクリプタがあるので受信処理へ。
	Rx(q);

	rx_packet.Clear();
	event.time = 500_usec;
	scheduler->RestartEvent(event);
}

// 受信。(rx_packet が埋まっていてかつ空きディスクリプタがある状態で呼ぶこと)
void
VirtIONetDevice::Rx(VirtQueue *q)
{
	VirtIONetReq req;
	VirtQDesc desc;
	uint32 desc_idx;
	uint32 nseg = 0;	// ゲストが指定したセグメント数
	uint32 buflen = 0;	// ゲストが指定したバッファのバイト数
	uint32 datalen;		// 実際に書き込んだバイト数

	assert(rx_packet.length != 0);
	assert(q->last_avail_idx != ReadLE16(q->driver + 2));

	// XXX どこでやるのがいいか
	// VirtIO は FCS を含まない 1514 バイトが上限と決まっているが、
	// HostNet の下の受信ドライバが FCS を含めているかどうかが分からない。
	// 仕方ないので 1514 バイトを超える時だけ取り除く…。
	if (rx_packet.length > 1514) {
		rx_packet.length = 1514;
	}

	desc_idx = ReadLE16(q->driver + 4 + (q->last_avail_idx % q->num) * 2);
	req.idx = desc_idx;
	do {
		buflen += ReadDesc(q, &desc, desc_idx);
		if (__predict_false(desc.IsWrite() == false)) {
			putlog(0, "Segments in rxq must be device-writeable");
			return;
		}
		nseg++;
		AddDataSegment(&req, desc.addr, desc.len, NULL);

		desc_idx = desc.next;
	} while (desc.IsNext());
	putlog(3, "buflen=%08x", buflen);

	memset(&req.hdr, 0, sizeof(req.hdr));
	req.hdr.hdr_len = htole16(sizeof(req.hdr));
	// XXX nseg は実際に使ったセグメントをカウントしないといけない
	req.hdr.num_buffers = htole16(nseg);

	// 書き込む。
	const uint8 *s = (const uint8 *)&req.hdr;
	for (int i = 0; i < sizeof(req.hdr); i++) {
		ReqAppendByte(&req, *s++);
	}
	datalen = sizeof(req.hdr);

	s = rx_packet.data();
	for (int i = 0; i < rx_packet.length; i++) {
		if (ReqAppendByte(&req, rx_packet[i]) == false) {
			// どうする?
			putlog(0, "buffer too small: buflen=%u rx_packet=%u",
				buflen, rx_packet.length);
		}
		datalen++;
	}

	CommitDesc(q, req.idx, datalen);

	// ここは VM スレッドなのでそのまま割り込みを上げる。
	Done(q);
}

const char *
VirtIONetDevice::GetFeatureName(int feature) const
{
	static std::pair<uint, const char *> names[] = {
		{ VIRTIO_NET_F_CSUM,			"CSUM" },
		{ VIRTIO_NET_F_GUEST_CSUM,		"G_CSUM" },
		{ VIRTIO_NET_F_CTRL_GUEST_OFFLOADS,	"C_G_OFFL" },
		{ VIRTIO_NET_F_MTU,				"MTU" },
		{ VIRTIO_NET_F_MAC,				"MAC" },
		{ VIRTIO_NET_F_GUEST_TSO4,		"G_TSO4" },
		{ VIRTIO_NET_F_GUEST_TSO6,		"G_TSO6" },
		{ VIRTIO_NET_F_GUEST_ECN,		"G_ECN" },
		{ VIRTIO_NET_F_GUEST_UFO,		"G_UFO" },
		{ VIRTIO_NET_F_HOST_TSO4,		"H_TSO4" },
		{ VIRTIO_NET_F_HOST_TSO6,		"H_TSO6" },
		{ VIRTIO_NET_F_HOST_ECN,		"H_ECN" },
		{ VIRTIO_NET_F_HOST_UFO,		"H_UFO" },
		{ VIRTIO_NET_F_MRG_RXBUF,		"MRG_RX" },
		{ VIRTIO_NET_F_STATUS,			"STATUS" },
		{ VIRTIO_NET_F_CTRL_VQ,			"CTL_VQ" },
		{ VIRTIO_NET_F_CTRL_RX,			"CTL_RX" },
		{ VIRTIO_NET_F_CTRL_VLAN,		"CTL_VLAN" },
		{ VIRTIO_NET_F_GUEST_ANNOUNCE,	"G_ANN" },
		{ VIRTIO_NET_F_MQ,				"MQ" },
		{ VIRTIO_NET_F_CTRL_MAC_ADDR,	"CTL_MAC" },
		{ VIRTIO_NET_F_RSC_EXT,			"RSC_EXT" },
		{ VIRTIO_NET_F_STANDBY,			"STANDBY" },
	};

	for (auto& p : names) {
		if (feature == p.first) {
			return p.second;
		}
	}
	return inherited::GetFeatureName(feature);
}

// デバイス構成レイアウトを書き出す。
void
VirtIONetConfigWriter::WriteTo(uint8 *dst) const
{
	MemoryStreamLE mem(dst);

	for (auto c : mac) {
		mem.Write8(c);
	}
	mem.Write16(status);
	mem.Write16(max_virtqueue_pairs);
	mem.Write16(mtu);
}
